#include "mainwindow.h"
#include "ui_mainwindow.h"

#include <QFileDialog>
#include <QThread>
#include <QTimer>

#include "wrappers.h"
#include "trackerworker.h"
#include "controlworker.h"

MainWindow::MainWindow(QWidget *parent) :
    QMainWindow(parent),
    ui(new Ui::MainWindow),
    backendPid(0),
    backendPipe(-1)
{
    ui->setupUi(this);

    /* Create a thread for workers */
    QThread* workerThread = new QThread(this);

    /* Create a worker to update the tracker */
    TrackerWorker * trackerWorker = new TrackerWorker();

    /* Move it to the worker thread. This means that slots of this worker
     * will run in the worker thread, and not block the UI */
    trackerWorker->moveToThread(workerThread);

    /* Connect tracker worker */
    connect(trackerWorker, SIGNAL (finished(float, float, float, float, float, float)),
            this, SLOT (updateTracker(float, float, float, float, float, float)));

    /* Create another worker for the control graph */
    QThread * cwThread = new QThread(this);
    ControlWorker * controlWorker = new ControlWorker();
    controlWorker->moveToThread(cwThread);

    /* Connect signals from control worker */
    connect(controlWorker, SIGNAL (gotNodes(QStringList)), this, SLOT (newNodes(QStringList)));
    connect(controlWorker, SIGNAL (gotParams(QStringList)), this, SLOT (newParams(QStringList)));
    connect(controlWorker, SIGNAL (gotParamValue(float)), this, SLOT (newParamValue(float)));
    connect(controlWorker, SIGNAL (gotConstantBlocks(QStringList)), this, SLOT (newConstantBlocks(QStringList)));
    connect(controlWorker, SIGNAL (paramSet(QString, QString)), controlWorker, SLOT (getParamValue(QString, QString)));

    /* Signals to control worker */
    connect(findChild<QPushButton *>("pbControlRefresh"), SIGNAL (clicked()), controlWorker, SLOT (getNodes()));
    connect(findChild<QComboBox *>("nodeSelect"), SIGNAL (currentIndexChanged(QString)), controlWorker, SLOT (getParams(QString)));
    connect(this, SIGNAL (getParamValue(QString, QString)), controlWorker, SLOT (getParamValue(QString, QString)));
    connect(this, SIGNAL (setParamValue(QString, QString, float)), controlWorker, SLOT (setParamValue(QString, QString, float)));

    /* Connect and disconnect from backend when signals emitted */
    connect(this, SIGNAL (connectWorkers()), trackerWorker, SLOT (connectBackend()));
    connect(this, SIGNAL (disconnectWorkers()), trackerWorker, SLOT (disconnectBackend()));
    connect(this, SIGNAL (connectWorkers()), controlWorker, SLOT (connectBackend()));
    connect(this, SIGNAL (disconnectWorkers()), controlWorker, SLOT (disconnectBackend()));

    /* Connect refresh button and refresh timer to tracker worker */
    QTimer * trackerTimer = new QTimer(this);
    connect(trackerTimer, SIGNAL(timeout()), trackerWorker, SLOT(process()));
    connect(findChild<QPushButton *>("pbRefresh"), SIGNAL (clicked()), trackerWorker, SLOT (process()));

    /* Start the things */
    trackerTimer->start(300);
    workerThread->start();
    cwThread->start();
}

MainWindow::~MainWindow()
{
    delete ui;
}

void MainWindow::updateConsole()
{
    if (backendPipe != -1) {
        char buf[256];
        size_t len = 0;
        len = readBackend(backendPipe, buf, len);
        if (len > 0) {
            QLineEdit * con = findChild<QLineEdit *>("vConsole");
            con->setText(con->text().append(buf));
        }
    }
}

void MainWindow::updateTracker(float x, float y, float z, float p, float r, float yaw)
{
    findChild<QLineEdit *>("xLineEdit")->setText(QString::number(x));
    findChild<QLineEdit *>("yLineEdit")->setText(QString::number(y));
    findChild<QLineEdit *>("zLineEdit")->setText(QString::number(z));
    findChild<QLineEdit *>("pLineEdit")->setText(QString::number(p));
    findChild<QLineEdit *>("rLineEdit")->setText(QString::number(r));
    findChild<QLineEdit *>("yLineEdit_2")->setText(QString::number(yaw));
}

void MainWindow::on_pbStart_clicked()
{
    this->backendPid = startBackend(findChild<QLineEdit *>("backendPath")->text().toStdString().c_str(), &backendPipe);
    findChild<QPushButton *>("pbStart")->setEnabled(false);
    findChild<QPushButton *>("pbStop")->setEnabled(true);
    backendState = 1;
}

void MainWindow::on_pbConnect_clicked()
{
    findChild<QPushButton *>("pbStart")->setEnabled(false);
    findChild<QPushButton *>("pbConnect")->setEnabled(false);
    findChild<QPushButton *>("pbStop")->setEnabled(true);
    emit(connectWorkers());
    backendState = 1;
}

void MainWindow::on_pbStop_clicked()
{
    emit(disconnectWorkers());
    if (backendPid) {
        stopBackend(backendPid, backendPipe);
        backendPipe = -1;
        backendPid = 0;
    }
    findChild<QPushButton *>("pbStart")->setEnabled(true);
    findChild<QPushButton *>("pbConnect")->setEnabled(true);
    findChild<QPushButton *>("pbStop")->setEnabled(false);
    backendState = 0;
}

void MainWindow::on_chooseBackend_clicked()
{
    QString backendPath = QFileDialog::getOpenFileName(this,
         tr("Path to Backend Executable"));
    findChild<QLineEdit *>("backendPath")->setText(backendPath);
}


void MainWindow::newNodes(QStringList blocks)
{
    QComboBox * select = findChild<QComboBox *>("nodeSelect");
    select->clear(); /* Hopefully this clears the list */
    select->addItems(blocks);
}


void MainWindow::newConstantBlocks(QStringList blocks)
{
    QComboBox * xSelect = findChild<QComboBox *>("xSetpointSelect");
    xSelect->clear(); /* Hopefully this clears the list */
    xSelect->addItems(blocks);

    QComboBox * ySelect = findChild<QComboBox *>("ySetpointSelect");
    ySelect->clear(); /* Hopefully this clears the list */
    ySelect->addItems(blocks);

    QComboBox * zSelect = findChild<QComboBox *>("zSetpointSelect");
    zSelect->clear(); /* Hopefully this clears the list */
    zSelect->addItems(blocks);

    for (ssize_t i = 0; i < blocks.size(); i++) {
        if (blocks[i].contains("setpoint", Qt::CaseInsensitive) || blocks[i].contains("sp", Qt::CaseInsensitive)) {
            if (blocks[i].contains("x ", Qt::CaseInsensitive)) {
                xSelect->setCurrentIndex(i);
            }
            if (blocks[i].contains("y ", Qt::CaseInsensitive)) {
                ySelect->setCurrentIndex(i);
            }
            if (blocks[i].contains("z ", Qt::CaseInsensitive) || blocks[i].contains("alt", Qt::CaseInsensitive)) {
                zSelect->setCurrentIndex(i);
            }
        }
    }
}

void MainWindow::newParams(QStringList params)
{
    QComboBox * select = findChild<QComboBox *>("paramSelect");
    select->clear(); /* Hopefully this clears the list */
    select->addItems(params);
}

void MainWindow::newParamValue(float val)
{
    findChild<QLineEdit *>("paramValue")->setText(QString::number(val));
}

void MainWindow::on_paramSelect_currentIndexChanged(const QString &arg1)
{
    emit(getParamValue(findChild<QComboBox *>("nodeSelect")->currentText(), arg1));
}

void MainWindow::on_paramValue_returnPressed()
{
    emit (setParamValue(findChild<QComboBox *>("nodeSelect")->currentText(),
                        findChild<QComboBox *>("paramSelect")->currentText(),
                        findChild<QLineEdit *>("paramValue")->text().toFloat()));
}
